@@ -429,23 +429,33 @@
   `(error "Can't supply a value for :FILL-POINTER (~S) that is larger ~
            than the~A size of the vector (~S)" ,actual ,adjective ,max))
 
-(declaim (inline %save-displaced-array-backpointer))
+(declaim (inline %save-displaced-array-backpointer
+                 %save-displaced-new-array-backpointer))
 (defun %save-displaced-array-backpointer (array data)
   (flet ((purge (pointers)
            (remove-if (lambda (value)
                         (or (not value) (eq array value)))
                       pointers
                       :key #'weak-pointer-value)))
-    ;; Add backpointer to the new data vector if it has a header.
+    (let ((old-data (%array-data array)))
+      (unless (eq old-data data)
+        ;; Add backpointer to the new data vector if it has a header.
+        (when (array-header-p data)
+          (setf (%array-displaced-from data)
+                (cons (make-weak-pointer array)
+                      (purge (%array-displaced-from data)))))
+        ;; Remove old backpointer, if any.
+        (when (array-header-p old-data)
+          (setf (%array-displaced-from old-data)
+                (purge (%array-displaced-from old-data))))))))
+
+(defun %save-displaced-new-array-backpointer (array data)
+  (flet ((purge (pointers)
+           (remove-if-not #'weak-pointer-value pointers)))
     (when (array-header-p data)
       (setf (%array-displaced-from data)
             (cons (make-weak-pointer array)
-                  (purge (%array-displaced-from data)))))
-    ;; Remove old backpointer, if any.
-    (let ((old-data (%array-data array)))
-      (when (and (neq data old-data) (array-header-p old-data))
-        (setf (%array-displaced-from old-data)
-              (purge (%array-displaced-from old-data)))))))
+                  (purge (%array-displaced-from data)))))))
 
 (defmacro populate-dimensions (header list-or-index rank)
   `(if (listp ,list-or-index)
@@ -465,16 +475,13 @@
                (setq product (* product (the index dim)))
                (incf rank))))))
 
-(defconstant-eqx +widetag->element-type+
-    #.(let ((a (make-array 32 :initial-element 0)))
-        (dovector (saetp *specialized-array-element-type-properties* a)
-          (let ((tag (saetp-typecode saetp)))
-            (setf (aref a (ash (- tag #x80) -2)) (saetp-specifier saetp)))))
-  #'equalp)
-
 (declaim (inline widetag->element-type))
 (defun widetag->element-type (widetag)
-  (svref +widetag->element-type+ (- (ash widetag -2) 32)))
+  (svref #.(let ((a (make-array 32 :initial-element 0)))
+             (dovector (saetp *specialized-array-element-type-properties* a)
+               (let ((tag (saetp-typecode saetp)))
+                 (setf (aref a (ash (- tag #x80) -2)) (saetp-specifier saetp)))))
+         (- (ash widetag -2) 32)))
 
 (defun initial-contents-error (content-length length)
   (error "There are ~W elements in the :INITIAL-CONTENTS, but ~
@@ -521,7 +528,7 @@
                                                (vector-extra-data vector))
                                            (ash (sap-ref-word (current-fp) n-word-bytes) 3)) ; XXX: magic
                     (cond ((= widetag simple-vector-widetag)
-                           (fill vector (%make-lisp-obj no-tls-value-marker-widetag)))
+                           (fill vector (%make-lisp-obj unwritten-vector-element-marker)))
                           ((array-may-contain-random-bits-p widetag)
                            ;; Leave the last word alone for base-string,
                            ;; in case the mandatory trailing null is part of a data word.
@@ -556,13 +563,9 @@
              (setf (%array-data array) data)
              (setf (%array-displaced-from array) nil)
              (cond (displaced-to
-                    (let ((offset (or displaced-index-offset 0)))
-                      (when (> (+ offset total-size)
-                               (array-total-size displaced-to))
-                        (error "~S doesn't have enough elements." displaced-to))
-                      (setf (%array-displacement array) offset)
-                      (setf (%array-displaced-p array) t)
-                      (%save-displaced-array-backpointer array data)))
+                    (setf (%array-displacement array) (or displaced-index-offset 0))
+                    (setf (%array-displaced-p array) t)
+                    (%save-displaced-new-array-backpointer array data))
                    (t
                     (setf (%array-displaced-p array) nil)))
              (populate-dimensions array dimensions array-rank)
@@ -736,19 +739,8 @@ of specialized arrays is supported."
                ,@decls
                (tagbody ,@forms))))))))
 
-;;; We need this constant to be dumped in such a way that genesis can directly
-;;; compute the symbol-value in cold load, and not defer the reference until
-;;; *!cold-toplevels* are evaluated. That's because GROW-HASH-TABLE calls REPLACE
-;;; which calls VECTOR-REPLACE which tries to reference the array of functions,
-;;; which had better not be deferred, or all hell breaks loose. It's actually
-;;; tricky to call any function in any file that has its toplevel forms evaluated
-;;; later than the current file, because you don't generally know whether code
-;;; blobs in that other file have had their L-T-V fixups patched in.
-(defconstant-eqx +blt-copier-for-widetag+ #.(make-array 32 :initial-element nil)
-  #'equalp)
-
 (macrolet ((%ref (accessor-getter extra-params)
-             `(funcall (,accessor-getter array) array index ,@extra-params))
+             `(sb-c::%funcall-no-nargs (,accessor-getter array) array index ,@extra-params))
            (define (accessor-name slow-accessor-name accessor-getter
                                   extra-params check-bounds)
              `(progn
@@ -764,7 +756,8 @@ of specialized arrays is supported."
                                      (safety 0)))
                   (%ref ,accessor-getter ,extra-params))
                 (defun ,slow-accessor-name (array index ,@extra-params)
-                  (declare (optimize speed (safety 0)))
+                  (declare (optimize speed (safety 0))
+                           (array array))
                   (if (not (%array-displaced-p array))
                       ;; The reasonably quick path of non-displaced complex
                       ;; arrays.
@@ -772,14 +765,14 @@ of specialized arrays is supported."
                         (%ref ,accessor-getter ,extra-params))
                       ;; The real slow path.
                       (with-array-data
-                          ((vector array)
+                          ((array array)
                            (index (locally
                                       (declare (optimize (speed 1) (safety 1)))
                                     (,@check-bounds index)))
                            (end)
                            :force-inline t)
                         (declare (ignore end))
-                        (,accessor-name vector index ,@extra-params)))))))
+                        (%ref ,accessor-getter ,extra-params)))))))
   (define hairy-data-vector-ref slow-hairy-data-vector-ref
     %find-data-vector-reffer
     nil (progn))
@@ -796,7 +789,8 @@ of specialized arrays is supported."
     (new-value) (check-bound array (%array-dimension array 0))))
 
 (defun hairy-ref-error (array index &optional new-value)
-  (declare (ignore index new-value))
+  (declare (ignore index new-value)
+           (optimize (sb-c:verify-arg-count 0)))
   (error 'type-error
          :datum array
          :expected-type 'vector))
@@ -812,9 +806,9 @@ of specialized arrays is supported."
                            (ignorable index))
                   ,(if type
                        `(data-vector-ref (the ,atype vector)
-                                         (locally
-                                             (declare (optimize (safety 1)))
-                                           (the index
+                                         (the index
+                                              (locally
+                                                  (declare (optimize (safety 1)))
                                                 (,@check-form index))))
                        `(data-nil-vector-ref (the ,atype vector) index)))))
            (define-setter (saetp check-form)
@@ -858,7 +852,6 @@ of specialized arrays is supported."
                         collect `(setf (svref ,symbol ,widetag)
                                        (,deffer ,saetp ,check-form))))))
   (defun !hairy-data-vector-reffer-init ()
-    (!blt-copiers-cold-init +blt-copier-for-widetag+)
     (define-reffers %%data-vector-reffers%% define-reffer
       (progn)
       #'slow-hairy-data-vector-ref)
@@ -1043,24 +1036,17 @@ of specialized arrays is supported."
 
 ;;;; miscellaneous array properties
 
-(macrolet ((widetag->type (accessor)
-             (let ((table
-                    ;; Using unbound-marker for empty elements would be preferable,
-                    ;; but I'd either need a cross-compiler proxy for unbound-marker,
-                    ;; or else wrap a LOAD-TIME-VALUE on the table construction form
-                    ;; which might introduce more order-sensitivity to self-build.
-                    (sb-xc:make-array 32 ;:initial-element (make-unbound-marker)
-                                      :initial-element 0)))
-               (dovector (saetp *specialized-array-element-type-properties*)
-                 (setf (aref table (ash (- (saetp-typecode saetp) 128) -2))
-                       (funcall accessor saetp)))
-               `(aref ,table (- (ash (array-underlying-widetag array) -2) 32)))))
+(define-load-time-global *saetp-widetag-ctype* (make-array 32 :initial-element (make-unbound-marker)))
+
 (defun array-element-ctype (array)
   ;; same as (SPECIFIER-TYPE (ARRAY-ELEMENT-TYPE ARRAY)) but more efficient
-  (widetag->type saetp-ctype))
+  (svref *saetp-widetag-ctype*
+         (- (ash (array-underlying-widetag array) -2) 32)))
+
 (defun array-element-type (array)
   "Return the type of the elements of the array"
-  (truly-the (or list symbol) (widetag->type saetp-specifier))))
+  (truly-the (or list symbol)
+             (widetag->element-type (array-underlying-widetag array))))
 
 (defun array-rank (array)
   "Return the number of dimensions of ARRAY."
@@ -1137,26 +1123,26 @@ of specialized arrays is supported."
          :format-control "~S is not an array with a fill pointer."
          :format-arguments (list vector)))
 
-(declaim (inline fill-pointer))
+
 (defun fill-pointer (vector)
   "Return the FILL-POINTER of the given VECTOR."
   (declare (explicit-check))
-  (if (array-has-fill-pointer-p vector)
-      (%array-fill-pointer vector)
-      (fill-pointer-error vector)))
+  (fill-pointer vector))
 
 (defun %set-fill-pointer (vector new)
   (declare (explicit-check))
-  (cond ((not (array-has-fill-pointer-p vector)) (fill-pointer-error vector))
-        ((> (the index new) (%array-available-elements vector))
-         (let ((max (%array-available-elements vector)))
-           (error 'simple-type-error
-                  :datum new
-                  :expected-type (list 'integer 0 max)
-                  :format-control "The new fill pointer, ~S, is larger than the length of the vector (~S.)"
-                  :format-arguments (list new max))))
+  (cond ((not (and (arrayp vector)
+                   (array-has-fill-pointer-p vector)))
+         (fill-pointer-error vector))
         (t
-         (setf (%array-fill-pointer vector) new))))
+         (let ((max (%array-available-elements vector)))
+           (when (> (the fixnum new) max)
+             (error 'simple-type-error
+                    :datum new
+                    :expected-type (list 'integer 0 max)
+                    :format-control "The new fill pointer, ~S, is larger than the length of the vector (~S.)"
+                    :format-arguments (list new max)))
+           (setf (%array-fill-pointer vector) (truly-the index new))))))
 
 ;;; FIXME: It'd probably make sense to use a MACROLET to share the
 ;;; guts of VECTOR-PUSH between VECTOR-PUSH-EXTEND. Such a macro
@@ -1178,20 +1164,6 @@ of specialized arrays is supported."
            (setf (%array-fill-pointer array) (1+ fill-pointer))
            fill-pointer))))
 
-(defun !blt-copiers-cold-init (array)
-  (macrolet ((init ()
-               `(progn
-                  ,@(loop for saetp across *specialized-array-element-type-properties*
-                          when (and (not (member (saetp-specifier saetp) '(t nil)))
-                                    (<= (saetp-n-bits saetp) n-word-bits))
-                            collect `(setf (svref array ,(ash (- (saetp-typecode saetp) 128) -2))
-                                           #',(intern (format nil "UB~D-BASH-COPY"
-                                                              (saetp-n-bits saetp))
-                                                      "SB-KERNEL"))))))
-      (init)))
-
-(defmacro blt-copier-for-widetag (x) `(aref +blt-copier-for-widetag+ (ash (- ,x 128) -2)))
-
 (defun extend-vector (vector min-extension)
   (declare (optimize speed)
            (vector vector))
@@ -1217,8 +1189,8 @@ of specialized arrays is supported."
                      (truly-the simple-vector old-data)
                      :start2 old-start :end2 old-end)
             (let ((copier (blt-copier-for-widetag widetag)))
-              (if (functionp copier)
-                  (funcall copier old-data old-start new-data 0 old-length)
+              (if copier
+                  (funcall (truly-the function copier) old-data old-start new-data 0 old-length)
                   (replace new-data old-data :start2 old-start :end2 old-end))))
         (setf (%array-data vector) new-data
               (%array-available-elements vector) new-length
@@ -1384,15 +1356,38 @@ of specialized arrays is supported."
 ;;; dangerous to do so: shrinking the size of an object accessible
 ;;; to another thread could cause it to access an out-of-bounds element.
 ;;; GC should generally be fine no matter what happens, because it either
-;;; reads the old length or the new length. If the old, and unboxed,
-;;; the whole vector is skipped; if simple-vector, then at worst it reads
-;;; the header word for a filler, which is a valid element (yup really!).
-;;; If it reads the new length, then the next object is filler.
-(defun %shrink-vector (vector new-length)
+;;; reads the old length or the new length. If it reads the old length,
+;;; then the whole vector is skipped if unboxed; if it reads the new length,
+;;; then the next object is a filler.
+;;; Exception: for SIMPLE-VECTOR we always zeroized the unused tail,
+;;; because the garbage collector can scan certain pages without regard
+;;; to object boundaries. The situation we need to avoid is this:
+;;;       "old" #(...............|.....)
+;;;       "new" #(..........)Fill|.....
+;;;                              ^ page boundary
+;;; where GC reads the objects on the page just after the filler
+;;; because it doesn't know not to.
+;;;
+(defmacro make-filler (n)
+  `(logior (ash ,n #+64-bit 32 #-64-bit ,n-widetag-bits) filler-widetag))
+(defmacro filler-nwords (header)
+  `(ash ,header #+64-bit -32 #-64-bit ,(- n-widetag-bits)))
+
+(defun %shrink-vector (vector new-length
+                              &aux (old-length (length vector))
+                                   (new-length* new-length))
   (declare (vector vector))
-  (unless (or (array-header-p vector) (typep vector '(simple-array nil (*))))
-    (let ((old-length (length vector))
-          (new-length new-length))
+  (cond
+    ((simple-vector-p vector)
+     ;; We do in fact call %SHRINK-VECTOR a lot from sequence functions
+     ;; that overallocate a temporary result. In all places where that happens,
+     ;; the discarded suffix was never used. So assuming pre-zeroed heap,
+     ;; it kind of just worked. But I don't want to assume that.
+     ;; For what it's worth, adding this assertion prior to FILL:
+     ;;   (WHEN (FIND 0 VECTOR :START OLD-LENGTH :TEST #'NEQ) (BUG "No can do"))
+     ;; produced no failures in the regression suite.
+     (when (< new-length old-length) (fill vector 0 :start new-length)))
+    ((not (or (array-header-p vector) (typep vector '(simple-array nil (*)))))
       (when (simple-base-string-p vector)
         ;; We can blindly store the hidden #\null at NEW-LENGTH, but it would
         ;; appear to be an out-of-bounds access if the length is not
@@ -1403,14 +1398,11 @@ of specialized arrays is supported."
         ;; Now treat both the old and new lengths as if they include
         ;; the byte that holds the implicit string terminator.
         (incf old-length)
-        (incf new-length))
+        (incf new-length*))
       (let* ((n-bits-shift (aref %%simple-array-n-bits-shifts%%
                                  (%other-pointer-widetag vector)))
              (old-nwords (ceiling (ash old-length n-bits-shift) n-word-bits))
-             (new-nwords (ceiling (ash new-length n-bits-shift) n-word-bits)))
-        ;; If we want to impose a constraint that unused bytes above the
-        ;; new length and below the physical end are 0,
-        ;; then now would be the time to enforce that.
+             (new-nwords (ceiling (ash new-length* n-bits-shift) n-word-bits)))
         (when (< new-nwords old-nwords)
           (with-pinned-objects (vector)
             ;; VECTOR-SAP is only for unboxed vectors. Use the vop directly.
@@ -1430,8 +1422,7 @@ of specialized arrays is supported."
                     ;; if GC sees this bit pattern prior to setting the new length;
                     ;; but even for SIMPLE-VECTOR, it's OK, it turns out.
                     (setf (sap-ref-word data (ash aligned-new word-shift))
-                          (logior (ash (1- diff) n-widetag-bits)
-                                  filler-widetag)))))))))))
+                          (make-filler diff)))))))))))
   ;; Only arrays have fill-pointers, but vectors have their length
   ;; parameter in the same place.
   (setf (%array-fill-pointer vector) new-length)
@@ -1906,5 +1897,4 @@ function to be removed without further warning."
 
 (defun weak-vector-p (x)
   (and (simple-vector-p x)
-       #+(or x86 x86-64) (test-header-data-bit x (ash vector-weak-flag array-flags-data-position))
-       #-(or x86 x86-64) (logtest (get-header-data x) (ash vector-weak-flag array-flags-data-position))))
+       (test-header-data-bit x (ash vector-weak-flag array-flags-data-position))))
